
/*
 * splinter 0.1.0
 * 2016, 2017
 * published under BSD license, see license.txt
 * by Graeme Sheppard www.rillion.net/splinter/
 * a reworking and extension of jumpgate by
 * Patroklos Argyroudis http://jumpgate.sourceforge.net/
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>


#define nMaxIPchars       20    /* length > 111.222.111.222 */
#define nMaxFD            23    /* above the number id of network FD - unsure how this was arrived at */
#define nTrafficSize    4096    /* length of network data exchange */
#define nBacklogSize    1024    /* size of pending queue */
#define nMaxLine         256    /* length of line of text */


typedef enum {eCount, eLabels, eFinal} eCfgPass;
typedef enum {eEmpty = 0, eSingle, eCidr, eRange} eMaskType;
const char szMasks[] = "0scr"; /* in order as eMaskType above */


struct splinter_record
{
    eMaskType iMaskType;
    unsigned int iFrom;
    unsigned int iTo;
    in_addr_t iDestAddr;
    int iDestPort;
};


struct label_record
{
    char* szLabelName;
    in_addr_t iDestAddr;
    int iDestPort;
};
    

struct config_data
{
    const char* szFileName;
    struct splinter_record* pSplinters;
    struct label_record* pLabelData;
    char* pLabelNames;
    int iSplinters;
    int iLabelCount;
    int iLabelChars;
    in_addr_t iListenAddr;
    int iListenPort;
    in_addr_t iDefaultAddr;
    int iDefaultPort;
};


struct jumpgate_data
{
    struct in_addr iSourceAddr;
    int iFD;
    in_addr_t iServerAddr;
    int iServerPort;
    struct timeval sStartTime;
    int iTCPerrors;
};


inline void wipe_string (char *szText)
{
    szText [0] = '\000';
}


void log_message (const char* szFlags, const char* szMessage)
{
    const char szLogLevel[] = "loglevel";
    const char szSysLog[] = "syslog";
    static int iLogLevel = 0;
    static int iSysLog = 0;
    
    char szOut [nMaxLine];
    int iErrorNum, iLevel;
    const char* pLetter;
    
    if (strncmp (szMessage, szLogLevel, sizeof (szLogLevel)) == 0)
    {
        iLogLevel = atoi (szFlags);
        return;
    }
    
    if (strncmp (szMessage, szSysLog, sizeof (szSysLog)) == 0)
    {
        iSysLog = 1;
        return;
    }
    
    wipe_string (szOut);
    iErrorNum = errno;
    if (!strchr (szFlags, 'f'))
        iErrorNum = 0;
    
    if (strchr (szFlags, 't'))
    {
        time_t tTime;
        struct tm *pTime;
        
        tTime = time (NULL);
        pTime = localtime (&tTime);
        strftime (szOut, sizeof (szOut), "%Y-%m-%d %H:%M:%S ", pTime);
    }
    
    strncat (szOut, szMessage, sizeof (szOut));
    
    if (iErrorNum > 0)
    {
        strncat (szOut, ": ", sizeof (szOut));
        strncat (szOut, strerror (iErrorNum), sizeof (szOut));
    }
    
    iLevel = 0;
    pLetter = &szFlags [0];
    while (*pLetter != '\000')
    {
        if (*pLetter >= '0' && *pLetter <= '9')
        {
            iLevel = *pLetter - '0';
            break;
        }
        pLetter++;
    }

    if (strchr (szFlags, 'e'))
        fprintf (stderr, "%s\n", szOut);
    else
    {
        if (iLevel <= iLogLevel)
        {
            printf ("%s\n", szOut);
        }
    }

    if (iSysLog)
      syslog (LOG_NOTICE, szMessage);   

    if (strchr (szFlags, 'x'))
        exit (EXIT_SUCCESS);
    if (strchr (szFlags, 'X'))
        exit (EXIT_FAILURE);
}


struct timeval micro_ticks()
{
    struct timeval sTime;
    
    if (gettimeofday (&sTime, NULL) == -1)
    {
        sTime.tv_sec = 0;
        sTime.tv_usec = 0;
    }
    
    return sTime;
}


void text_time_diff (struct timeval sStart, struct timeval sEnd, char* szText, int iSize)
{
    time_t iSeconds;
    suseconds_t iMicros;
    unsigned long iLongSeconds;
       
    if (sStart.tv_sec <= 0 || sEnd.tv_sec <= 0)
    {
        snprintf (szText, iSize, "time_err");
        return;
    }

    iSeconds = sEnd.tv_sec - sStart.tv_sec;
    if (sEnd.tv_usec >= sStart.tv_usec)
        iMicros = sEnd.tv_usec - sStart.tv_usec;
    else
    {
        iMicros = (sEnd.tv_usec + 1000000) - sStart.tv_usec;
        iSeconds--;
    }
    /* type casting */
    iLongSeconds = iSeconds;    
    snprintf (szText, iSize, "%lu.%03d", iLongSeconds, (int) iMicros / 1000);
}


unsigned int flip_ip (unsigned int iIn)
{
    unsigned char* pSource = (unsigned char*) &iIn;
    unsigned char sDest [4];
    unsigned int* pDest = (unsigned int*) &sDest [0];
    
    sDest [0] = pSource [3];
    sDest [1] = pSource [2];
    sDest [2] = pSource [1];
    sDest [3] = pSource [0];
    return *pDest;
}


struct in_addr ip_transform (in_addr_t iIP)
{
    struct in_addr sIP;

    sIP.s_addr = iIP;
    return sIP;
}


int splinter_filter (in_addr_t iSourceAddr, const struct config_data* pConfig, const struct splinter_record* pSplinters, in_addr_t *dest_addr, int *dest_port)
{
    struct splinter_record pF;
    eMaskType iType;
    int iFound = 0;
    unsigned int iIP;
    
    iIP = flip_ip ((unsigned int) iSourceAddr);
    
    for (;;)
    {
        pF = pSplinters [iFound];
        iType = pF.iMaskType;
        if (iType == eEmpty)
        {
            *dest_addr = pConfig->iDefaultAddr;
            *dest_port = pConfig->iDefaultPort;
            return -1;
        }
        if (iType == eSingle && iIP == pF.iFrom)
            break;
        if (iType == eCidr)
        {
            if (pF.iTo == 0) /* catch all */
                break;
            if ((iIP & pF.iTo) == pF.iFrom)
                break;
        }
        if (iType == eRange && iIP >= pF.iFrom && iIP <= pF.iTo)
            break;
        iFound++;
    }

    pF = pSplinters [iFound];
    *dest_addr = pF.iDestAddr;
    *dest_port = pF.iDestPort;
    return iFound;
}


void trim_spaces (char* szText)
{
    char szReduce [nMaxLine];
    int iLength, iIndex;
    char cLetter;
    
    iLength = strlen (szText);
    if (iLength == 0)
        return;

    iIndex = 0;
    for (;;)
    {
        cLetter = szText [iIndex];
        if (cLetter != ' ')
            break;
        iIndex++;
    }

    if (iIndex > 0)
    {
        snprintf (szReduce, nMaxLine, "%s", szText + iIndex);
        snprintf (szText, nMaxLine, "%s", szReduce);
    }

    iLength = strlen (szText);
    if (iLength == 0)
        return;

    iIndex = iLength - 1;
    for (;;)
    {
        if (iIndex < 0)
            break;
        cLetter = szText [iIndex];
        if (cLetter != ' ')
            break;
        iIndex--;
    }
    
    szText [iIndex + 1] = '\000';
}


unsigned int atoip (const char *szText)
{
    in_addr_t iAddr;
    int iReturnCode;

    iReturnCode = inet_pton (AF_INET, szText, &iAddr);
    if (iReturnCode != 1)
        return 0;
    else
        return flip_ip ((unsigned int) iAddr);
}


void socket_to_text (struct in_addr iAddr, int iPort, char* szText, int iSize)
{
    if (iPort == 0)
        snprintf (szText, iSize, "0");
    else
        snprintf (szText, iSize, "%s:%d", inet_ntoa (iAddr), iPort);
}


void split_delimiter (char cDelimiter, char *pcFound, char *szText, char *szLeft, char *szRight)
{
    char *pcLetter;
    
    pcLetter = index (szText, cDelimiter);
    if (pcLetter == NULL)
        return;

    *pcFound = cDelimiter;
    *pcLetter = '\000';
    snprintf (szLeft, nMaxLine, szText);
    snprintf (szRight, nMaxLine, pcLetter + 1);
    wipe_string (szText);
}


void fill_splinter (struct splinter_record* pSplinter, eMaskType iMask, int iFrom, int iTo, int iDestAddr, int iDestPort)
{
    pSplinter->iMaskType = iMask;
    pSplinter->iFrom = iFrom;
    pSplinter->iTo = iTo;
    pSplinter->iDestAddr = iDestAddr;
    pSplinter->iDestPort = iDestPort;
}


void zero_splinter (struct splinter_record* pSplinter)
{
    fill_splinter (pSplinter, eEmpty, 0, 0, 0, 0);
}


void cfg_file (eCfgPass iPass, struct config_data* pConfig)
{
    FILE *fCfg;
    char szLine [nMaxLine];
    char szName [nMaxLine];
    char szValue [nMaxLine];
    char szNameLeft [nMaxLine];
    char szNameRight [nMaxLine];
    char szValueLeft [nMaxLine];
    char szValueRight [nMaxLine];
    char szError [nMaxLine];
    int iNameLeft, iNameRight;
    in_addr_t iDest;
    int iPort;
    char cName, cValue;
    char *szIn;
    char *pcLetter;
    int iLine;
    int iErrors = 0;
    int iReturnCode;
    int iSplinterIndex = 0;
    int iNameLength;
    int iLabelIndex = 0;
    int iLabelStylus = 0;
    int iCIDR, iAdd, iCount;
    int iSysLogCount = 0;
    struct splinter_record* pSplinters = pConfig->pSplinters;
    struct splinter_record aSplinter;
    static char szSysLogName [nMaxLine];
    
    fCfg = fopen (pConfig->szFileName, "r");
    if (fCfg == NULL)
    {
        snprintf (szLine, sizeof (szLine), "cannot open file \"%s\"", pConfig->szFileName);
        log_message ("feX", szLine);
    }
    aSplinter.iMaskType = eEmpty; /* '0' = empty */
    iLine = 0;
    wipe_string (szError);

    for (;;)
    {
        if (strncmp (szError, "", nMaxLine) != 0)
        {
            printf ("line %d: %s\n", iLine, szError);
            iErrors = 1;
            wipe_string (szError);
        }
        
        if (aSplinter.iMaskType != eEmpty)
        {
            switch (iPass)
            {
                case eCount:
                    iSplinterIndex++;
                    break;
                
                case eLabels:
                    break;
                    
                case eFinal:
                    fill_splinter (&pSplinters [iSplinterIndex], aSplinter.iMaskType, aSplinter.iFrom, aSplinter.iTo, aSplinter.iDestAddr, aSplinter.iDestPort);
                    iSplinterIndex++;
                    break;
            }
        }

        zero_splinter (&aSplinter);
        iLine++;
        szIn = fgets (szLine, sizeof (szLine), fCfg);
        if (szIn == NULL)
            break;
        
        pcLetter = strchr (szLine, '\n');
        if (pcLetter)
            *pcLetter = '\000';
        pcLetter = strchr (szLine, '#');
        if (pcLetter)
            *pcLetter = '\000';

        for (;;)
        {
            pcLetter = strchr (szLine, '\t');
            if (pcLetter == NULL)
                break;
            *pcLetter = ' ';
        }

        trim_spaces (szLine);
        if (strlen (szLine) == 0)
            continue;
            
        pcLetter = index (szLine, '=');
        if (pcLetter == NULL)
        {
            snprintf (szError, nMaxLine, "no = setting");
            continue;
        }
        
        *pcLetter = '\000';
        snprintf (szName, nMaxLine, szLine);
        snprintf (szValue, nMaxLine, pcLetter + 1);
        wipe_string (szNameLeft);
        wipe_string (szNameRight);
        wipe_string (szValueLeft);
        wipe_string (szValueRight);
        cName = cValue = '.';

        split_delimiter ('/', &cName, szName, szNameLeft, szNameRight);
        split_delimiter ('-', &cName, szName, szNameLeft, szNameRight);
        split_delimiter ('!', &cName, szName, szNameLeft, szNameRight);
        split_delimiter (':', &cValue, szValue, szValueLeft, szValueRight);
        split_delimiter ('!', &cValue, szValue, szValueLeft, szValueRight);
        
        trim_spaces (szName);
        trim_spaces (szValue);
        trim_spaces (szNameLeft);
        trim_spaces (szNameRight);
        trim_spaces (szValueLeft);
        trim_spaces (szValueRight);
        
        iDest = 0;
        iPort = -1;

        if (cValue == ':')
        {
            if (strncmp (szName, "listen", nMaxLine) == 0 && strncmp (szValueLeft, "", nMaxLine) == 0)
                iDest = htonl (INADDR_ANY); /* all network interfaces */
            else
            {
                iReturnCode = inet_pton (AF_INET, szValueLeft, &iDest);
                if (iReturnCode != 1)
                {
                    snprintf (szError, nMaxLine, "bad ip address");
                    continue;
                }
            }
            iPort = atoi (szValueRight);
            if (iPort <= 0 || iPort >= 65536)
            {
                snprintf (szError, nMaxLine, "bad destination port");
                continue;
            }
            wipe_string (szValueLeft);
            wipe_string (szValueRight);
        }

        if (cValue == '!' && (iPass == eLabels || iPass == eFinal))
        {
            int iIndex = 0;
            int iLimit = iPass == eFinal ? pConfig->iLabelCount : iLabelIndex;
            for (;;)
            {
                if (iIndex >= iLimit)
                {
                    iIndex = -1;
                    break;
                }
                if (strncmp (pConfig->pLabelData [iIndex].szLabelName, szValueRight, sizeof (szValueRight)) == 0)
                    break;
                iIndex++;
            }
            if (iIndex < 0)
            {
                if (iPass == eFinal)
                {
                    snprintf (szError, nMaxLine, "label not found");
                    continue;
                }
                iPort = 0; /* grace for eLabels pass */
            }
            else
            {
                iDest = pConfig->pLabelData [iIndex].iDestAddr;
                iPort = pConfig->pLabelData [iIndex].iDestPort;
            }
        }

        if (strncmp (szName, "syslog", nMaxLine) == 0)
        {
            if (iSysLogCount > 0)
            {
                snprintf (szError, nMaxLine, "more than one syslog");
                continue;
            }
            iSysLogCount++;
            if (strncmp (szValue, "", nMaxLine) == 0)
            {
                snprintf (szError, nMaxLine, "no syslog value");
                continue;
            }
            
            if (iPass == eFinal)
            {
                snprintf (szSysLogName, nMaxLine, szValue);
                openlog (szSysLogName, 0, LOG_USER);
                log_message (NULL, "syslog");
            }
            continue;
        }

        if (strncmp (szValue, "0", nMaxLine) == 0)
            iPort = 0;

        if ((iPass == eLabels || iPass == eFinal) && iPort < 0)
        {
            snprintf (szError, nMaxLine, "bad destination");
            continue;
        }
                
        if (strncmp (szName, "listen", nMaxLine) == 0)
        {
            if (cValue == '!')
            {
                snprintf (szError, nMaxLine, "cannot listen on a label");
                continue;
            }
            pConfig->iListenAddr = iDest;
            pConfig->iListenPort = iPort;
            continue;
        }
        
        if (strncmp (szName, "default", nMaxLine) == 0)
        {
            pConfig->iDefaultAddr = iDest;
            pConfig->iDefaultPort = iPort;
            continue;
        }

        if (cName == '!')
        {
            iNameLength = strlen (szNameRight);
            if (iNameLength <= 0)
            {
                snprintf (szError, nMaxLine, "label has no characters");
                continue;
            }
            if (iPass == eCount)
            {
                pConfig->iLabelCount++;
                pConfig->iLabelChars += iNameLength + 1; /* add one for trailing 0x00 */
            }
            else if (iPass == eLabels)
            {
                if (iLabelIndex > pConfig->iLabelCount)
                {
                    snprintf (szError, nMaxLine, "label count exceeded");
                    continue;
                }
                if (iLabelStylus + iNameLength + 1 > pConfig->iLabelChars)
                {
                    snprintf (szError, nMaxLine, "label name overflow");
                    continue;
                }
                int iIndex;
                for (iIndex = 0; iIndex < iLabelIndex; iIndex++)
                {
                    if (strncmp (szNameRight, pConfig->pLabelData [iIndex].szLabelName, sizeof (szNameRight)) == 0)
                    {
                        snprintf (szError, nMaxLine, "label \"%s\" repeated", szNameRight);
                        continue;
                    }
                }
                memcpy (pConfig->pLabelNames + iLabelStylus, szNameRight, iNameLength + 1); /* memcpy */
                pConfig->pLabelData [iLabelIndex].szLabelName = pConfig->pLabelNames + iLabelStylus;
                pConfig->pLabelData [iLabelIndex].iDestAddr = iDest;
                pConfig->pLabelData [iLabelIndex].iDestPort = iPort;
                iLabelIndex++;
                iLabelStylus += iNameLength + 1;
            }
            continue;
        }

        if (cName == '.') /* just an IP address then */
        {
            iNameLeft = atoip (szName);
            if (iNameLeft != 0)
            {
                fill_splinter (&aSplinter, eSingle, iNameLeft, 0, iDest, iPort);
                continue;
            }
            else
            {
                snprintf (szError, nMaxLine, "bad IP address");
                continue;
            }
        }
        else if (cName == '/')
        {
            if (strncmp (szNameLeft, "0", nMaxLine) == 0 && strncmp (szNameRight, "0", nMaxLine) == 0)
            /* the catch all */
            {
                fill_splinter (&aSplinter, eCidr, 0, 0, iDest, iPort);
                continue;
            }
            iNameLeft = atoip (szNameLeft);
            iNameRight = atoip (szNameRight);
            if (iNameRight == 0)
            {
                iCIDR = atoi (szNameRight);
                if (iCIDR >= 1 && iCIDR <= 32)
                {
                    iNameRight = 0x80000000;
                    iAdd = 0x40000000;
                    for (iCount = 2; iCount <= iCIDR; iCount++)
                    {
                        iNameRight += iAdd;
                        iAdd = iAdd >> 1;
                    }
                }
            }
            if (iNameLeft == 0 || iNameRight == 0)
            {
                snprintf (szError, nMaxLine, "bad cidr expression");
                continue;
            }
            fill_splinter (&aSplinter, eCidr, iNameLeft, iNameRight, iDest, iPort);
            continue;
        }
        else if (cName == '-')
        {
            iNameLeft = atoip (szNameLeft);
            iNameRight = atoip (szNameRight);
            if (iNameLeft == 0 || iNameRight == 0 || iNameLeft >= iNameRight)
            {
                snprintf (szError, nMaxLine, "bad ip range");
                continue;
            }
            fill_splinter (&aSplinter, eRange, iNameLeft, iNameRight, iDest, iPort);
            continue;
        }
        
        snprintf (szError, nMaxLine, "general syntax error");
    }
    
    fclose (fCfg);
    if (iErrors)
    {
        snprintf (szLine, sizeof (szLine), "errors in file \"%s\"", pConfig->szFileName);
        log_message ("eX", szLine);
    }

    switch (iPass)
    {
        case eCount:
            pConfig->iSplinters = iSplinterIndex;
            break;
            
        case eLabels:
            break;
            
        case eFinal:
            zero_splinter (&pSplinters [iSplinterIndex]);
            break;
    }    
}


void tcp_exit (const struct jumpgate_data* pJumpData, const char* szSection, const char* szError)
{
    int iErrorNum = 0;
    char szPrintTime [30];
    char szMessage [nMaxLine];
    char szExtendedError [nMaxLine];
    char szSource [30];
    char szDest [30];
    struct in_addr sDestAddr;

    if (szError)
        iErrorNum = errno;

    text_time_diff (pJumpData->sStartTime, micro_ticks(), szPrintTime, sizeof (szPrintTime));
    strncpy (szSource, inet_ntoa (pJumpData->iSourceAddr), sizeof (szSource));
    sDestAddr.s_addr = pJumpData->iServerAddr;
    strncpy (szDest, inet_ntoa (sDestAddr), sizeof (szDest));
    snprintf (szMessage, sizeof (szMessage), "%s -> %s:%d (%ss)", szSource, szDest, pJumpData->iServerPort, szPrintTime);

    if (pJumpData->iTCPerrors)
    {
        if (szError)
        {
            if (iErrorNum > 0)
                snprintf (szExtendedError, sizeof (szExtendedError), " (%s: %s in %s)", szError, strerror (iErrorNum), szSection);
            else
                snprintf (szExtendedError, sizeof (szExtendedError), " (%s in %s)", szError, szSection);
        }
        else
            snprintf (szExtendedError, sizeof (szExtendedError), " (in %s)", szSection);
    }
    else
        wipe_string (szExtendedError);

    strncat (szMessage, szExtendedError, sizeof (szMessage));
    log_message ("t5x", szMessage);
}


void tcp_jumpgate (struct jumpgate_data sJumpData)
{
    int iReadChars, iServerFD;
    char buf [nTrafficSize];
    fd_set setMonitor1, setMonitor2;
    struct sockaddr_in sServer;

    if ((iServerFD = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        tcp_exit (&sJumpData, "socket", "socket acquire error");

    sServer.sin_family = AF_INET;
    sServer.sin_port = htons (sJumpData.iServerPort);
    memcpy (&sServer.sin_addr, &sJumpData.iServerAddr, sizeof (in_addr_t));

    if ((connect (iServerFD, (struct sockaddr *) &sServer, sizeof (sServer))) < 0)
        tcp_exit (&sJumpData, "connect", "connect error");

    for (;;)
    {
        FD_ZERO (&setMonitor1);
        FD_ZERO (&setMonitor2);
        FD_SET (sJumpData.iFD, &setMonitor1);
        FD_SET (sJumpData.iFD, &setMonitor2);
        FD_SET (iServerFD, &setMonitor1);
        FD_SET (iServerFD, &setMonitor2);

        if (select (nMaxFD, &setMonitor1, NULL, &setMonitor2, NULL) == -1)
        {
            if ((shutdown (iServerFD, 2)) == -1)
                tcp_exit (&sJumpData, "select", "shutdown error");
            if ((close (iServerFD)) == -1)
                tcp_exit (&sJumpData, "select", "close error");
            tcp_exit (&sJumpData, "select", NULL);
        }

        if (FD_ISSET (sJumpData.iFD, &setMonitor1) || FD_ISSET (sJumpData.iFD, &setMonitor2))
        {
            if ((iReadChars = read (sJumpData.iFD, buf, nTrafficSize)) <= 0)
            {
                if ((shutdown (iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "client read", "shutdown error");
                if ((close (iServerFD)) == -1)
                    tcp_exit (&sJumpData, "client read", "close error");
                tcp_exit (&sJumpData, "client read", NULL);
            }
            if ((write (iServerFD, buf, iReadChars)) <= 0)
            {
                if ((shutdown (iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "server write", "shutdown error");
                if ((close (iServerFD)) == -1)
                    tcp_exit (&sJumpData, "server write", "close error");
                tcp_exit (&sJumpData, "server write", NULL);
            }
        }
        else if (FD_ISSET (iServerFD, &setMonitor1) || FD_ISSET (iServerFD, &setMonitor2))
        {
            if ((iReadChars = read (iServerFD, buf, nTrafficSize)) < 0)
            {
                if ((shutdown (iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "server read", "shutdown error");
                if ((close (iServerFD)) == -1)
                    tcp_exit (&sJumpData, "server read", "close error");
                tcp_exit (&sJumpData, "server read", NULL);
            }
            if ((write (sJumpData.iFD, buf, iReadChars)) <= 0)
            {
                if ((shutdown (iServerFD, 2)) == -1)
                    tcp_exit (&sJumpData, "client write", "shutdown error");
                if ((close (iServerFD)) == -1)
                    tcp_exit (&sJumpData, "client write", "close error");
                tcp_exit (&sJumpData, "client write", NULL);
            }
        }
    }
}


void child_signal (int iSignal)
{
    /*
     * Wait for all dead processes.
     * We use a non-blocking call to be sure this signal handler will not
     * block if a child was cleaned up in another part of the program.
     */

    while (waitpid (-1, NULL, WNOHANG) > 0)
    {
        /* do nothing */;
    }
}


int daemonize()
{
    pid_t iPID;

    if ((signal (SIGTTOU, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGTTIN, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGTSTP, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((signal (SIGHUP, SIG_IGN)) == SIG_ERR)
        log_message ("feX", "daemonize signal error");

    if ((iPID = fork()) < 0)
        return -1;
    else if (iPID != 0)
        exit (EXIT_SUCCESS);

    if (setsid() == -1)
        log_message ("feX", "setsid error");

    umask (0);
    return 0;
}


void help()
{
    printf ("spilinter 0.1.0\n");
    printf ("www.rillion.net\n");
    printf ("-f <configuration.file> (mandatory)\n");
    printf ("-t <ip> test ip address\n");
    printf ("-l dump listing\n");
    printf ("-d daemonize\n");
    printf ("-e tcp error data\n");
}


int main (int argc, char *argv[])
{
    char iLetter;
    int iLocalFD;
    const int iInt = 1;
    pid_t iChildPID;
    socklen_t iSocketSize;
    struct sockaddr_in sLocalSocket, sClientSocket;
    int iConnectionFD;
    int iConfDone = 0;
    int iDumpCfg = 0;
    int iDaemonize = 0;
    int iTCPerrors = 0;
    struct config_data sConfig;
    struct in_addr iTestIP;
    struct sigaction sSigAction;

    if (argc <= 1)
    {
        help();
        exit (EXIT_SUCCESS);
    }
        
    iTestIP.s_addr = 0;

    while ((iLetter = getopt (argc, argv, "vldef:t:")) > 0 && iLetter < 128)
    {
        switch (iLetter)
        {
            case 'v':
                help();
                exit (EXIT_SUCCESS);
            case 'l':
                iDumpCfg = 1;
                break;
            case 'd':
                iDaemonize = 1;
                break;
            case 'e':
                iTCPerrors = 1;
                break;
            case 'f':
                if (iConfDone != 0)
                    log_message ("eX", "error: extra config file specified");
                sConfig.szFileName = optarg;
                sConfig.pSplinters = NULL;
                sConfig.pLabelData = NULL;
                sConfig.pLabelNames = NULL;
                sConfig.iSplinters = 0;
                sConfig.iLabelCount = 0;
                sConfig.iLabelChars = 0;
                sConfig.iDefaultPort = -1;
                cfg_file (eCount, &sConfig);
                sConfig.pSplinters = (struct splinter_record*) calloc (sConfig.iSplinters + 1, sizeof (struct splinter_record));
                if (sConfig.pSplinters == NULL)
                    log_message ("eX", "error alocating memory to splinters");
                if (sConfig.iLabelCount > 0)
                {
                    sConfig.pLabelData = calloc (sConfig.iLabelCount, sizeof (struct label_record));
                    if (sConfig.pLabelData == NULL)
                        log_message ("eX", "error alocating memory to label data");
                    sConfig.pLabelNames = calloc (1, sConfig.iLabelChars);
                    if (sConfig.pLabelNames == NULL)
                        log_message ("eX", "error alocating memory to label names");
                }
                cfg_file (eLabels, &sConfig);
                cfg_file (eFinal, &sConfig);
                iConfDone = 1;
                break;
            case 't':
                if (iTestIP.s_addr != 0)
                    log_message ("eX", "error: may only test one IP");
                iTestIP.s_addr = flip_ip (atoip (optarg));
                if (iTestIP.s_addr == 0)
                    log_message ("eX", "error: bad IP to test");
                iDumpCfg = 1;
                break;
        }
    }

    if (iConfDone == 0)
        log_message ("eX", "no configuration file specified");
    if (sConfig.iListenPort == 0)
        log_message ("eX", "no listen port specified");
    if (sConfig.iListenPort < 1024 && getuid() != 0)
        log_message ("eX", "error: you're trying to listen on a low port without having root privileges");
    if (sConfig.iDefaultPort < 0)
        log_message ("eX", "no default destination specified");

    if (iDumpCfg)
    {
        struct in_addr iAddr;
        
        iAddr.s_addr = sConfig.iListenAddr;
        printf ("Listen: %s:%d\n", inet_ntoa (iAddr), sConfig.iListenPort);
        iAddr.s_addr = sConfig.iDefaultAddr;
        printf ("Default: %s:%d\n", inet_ntoa (iAddr), sConfig.iDefaultPort);
        char szDest [nMaxLine];
        int iIndex;
        for (iIndex = 0; iIndex < sConfig.iLabelCount; iIndex++)
        {
            iAddr.s_addr = sConfig.pLabelData [iIndex].iDestAddr;
            socket_to_text (iAddr, sConfig.pLabelData [iIndex].iDestPort, szDest, sizeof (szDest));
            printf ("!%s = %s\n", sConfig.pLabelData [iIndex].szLabelName, szDest);
        }
        const struct splinter_record *pF;
        struct in_addr iAddrStart;
        struct in_addr iAddrEnd;
        struct in_addr iAddrDest;
        char szIncoming [nMaxLine];
        char szRangeEnd [nMaxLine];
        for (iIndex = 0; iIndex < sConfig.iSplinters; iIndex++)
        {
            pF = &sConfig.pSplinters [iIndex];
            iAddrStart.s_addr = flip_ip (pF->iFrom);
            iAddrEnd.s_addr = flip_ip (pF->iTo);
            iAddrDest.s_addr = pF->iDestAddr;
            switch (pF->iMaskType)
            {
                case eSingle:
                    snprintf (szIncoming, sizeof (szIncoming), "%s", inet_ntoa (iAddrStart));
                    break;
                case eCidr:
                    if (pF->iTo == 0)
                        snprintf (szIncoming, sizeof (szIncoming), "0/0");
                    else
                        snprintf (szIncoming, sizeof (szIncoming), "%s/%08x", inet_ntoa (iAddrStart), pF->iTo);
                    break;
                case eRange:
                    snprintf (szRangeEnd, sizeof (szRangeEnd), "%s", inet_ntoa (iAddrEnd));
                    snprintf (szIncoming, sizeof (szIncoming), "%s-%s", inet_ntoa (iAddrStart), szRangeEnd);
                    break;
                default:
                    snprintf (szIncoming, sizeof (szIncoming), "?");
                    break;
            }
            socket_to_text (iAddrDest, pF->iDestPort, szDest, sizeof (szDest));
            printf ("%d: %s -> %s\n", iIndex, szIncoming, szDest);
        }

        if (iTestIP.s_addr != 0)
        {
            char szMatch [nMaxLine];
            in_addr_t iDest;
            int iPort;
            struct in_addr sDest;
            int iMatch = splinter_filter (iTestIP.s_addr, &sConfig, sConfig.pSplinters, &iDest, &iPort);
            if (iMatch < 0)
                snprintf (szMatch, sizeof (szMatch), "default");
            else
                snprintf (szMatch, sizeof (szMatch), "%d", iMatch);
            sDest.s_addr = iDest;
            socket_to_text (sDest, iPort, szDest, sizeof (szDest));
            printf ("IP test match=%s: %s -> %s\n", szMatch, inet_ntoa (iTestIP), szDest);
        }

        exit (0);
    }

    if (sConfig.iLabelCount > 0)
    {
        free (sConfig.pLabelData);
        free (sConfig.pLabelNames);
    }

    if (iDaemonize)
    {
        if (daemonize() == -1)
            log_message ("eX", "fork error");
        log_message ("0", "loglevel");
    }
    else
        log_message ("5", "loglevel");

    if ((iLocalFD = socket (AF_INET, SOCK_STREAM, 0)) < 0)
        log_message ("feX", "socket error");

    bzero (&sLocalSocket, sizeof (sLocalSocket));
    sLocalSocket.sin_family = AF_INET;
    sLocalSocket.sin_addr.s_addr = sConfig.iListenAddr;
    sLocalSocket.sin_port = htons (sConfig.iListenPort);

    /* if you use linux beware of the SO_REUSEADDR vulnerability,
     * although I think new linux versions fix this problem
     */
    if (setsockopt (iLocalFD, SOL_SOCKET, SO_REUSEADDR, &iInt, sizeof (iInt)) < 0)
        log_message ("feX", "setsockopt error");

    if (bind (iLocalFD, (struct sockaddr *) &sLocalSocket, sizeof (sLocalSocket)) < 0)
        log_message ("feX", "listen error");

    if (listen (iLocalFD, nBacklogSize) < 0)
        log_message ("feX", "listen error");

    memset (&sSigAction, 0, sizeof (sSigAction));
    sSigAction.sa_handler = child_signal;
    if (sigaction (SIGCHLD, &sSigAction, 0))
        log_message ("feX", "signal error");

    for (;;)
    {
        iSocketSize = sizeof (sClientSocket);
        if ((iConnectionFD = accept (iLocalFD, (struct sockaddr *) &sClientSocket, &iSocketSize)) < 0)
        {
            if (errno == EINTR || errno == ECONNABORTED)
                continue;
            else
                log_message ("feX", "accept connection error");
        }

        if ((iChildPID = fork()) == -1)
            log_message ("feX", "fork error");

        if (iChildPID == 0)
        {
            struct jumpgate_data sJumpData;
            in_addr_t sServerAddr;
            int iServerPort;
            char szLogMessage [nMaxLine];
            char szClientIP [nMaxIPchars];
            char szServerIP [nMaxIPchars];

            if ((close (iLocalFD)) == -1)
                log_message ("feX", "close error"); /* might be logging only though */

            splinter_filter (sClientSocket.sin_addr.s_addr, &sConfig, sConfig.pSplinters, &sServerAddr, &iServerPort);
            strncpy (szClientIP, inet_ntoa (sClientSocket.sin_addr), sizeof (szClientIP));

            sJumpData.iSourceAddr = sClientSocket.sin_addr;
            sJumpData.iFD = iConnectionFD;
            sJumpData.iServerAddr = sServerAddr;
            sJumpData.iServerPort = iServerPort;
            sJumpData.sStartTime = micro_ticks();
            sJumpData.iTCPerrors = iTCPerrors;

            if (iServerPort <= 0) /* tear down connection */
            {
                snprintf (szLogMessage, nMaxLine, "%s -> 0", szClientIP);
                log_message ("t5", szLogMessage);
                if ((shutdown (iConnectionFD, 2)) == -1)
                    tcp_exit (&sJumpData, "client teardown", "shutdown error");
                if ((close (iConnectionFD)) == -1)
                    tcp_exit (&sJumpData, "client teardown", "close error");
                exit (EXIT_SUCCESS);
            }

            strncpy (szServerIP, inet_ntoa (ip_transform (sServerAddr)), sizeof (szServerIP));
            snprintf (szLogMessage, nMaxLine, "%s -> %s:%d", szClientIP, szServerIP, iServerPort);
            log_message ("t5", szLogMessage);
            tcp_jumpgate (sJumpData);
            
            exit (EXIT_SUCCESS); /* added to be sure of child exit */
        }

        if ((close (iConnectionFD)) == -1)
            log_message ("feX", "close connection error"); /* this aborts too */
        
    }
    
    return EXIT_SUCCESS; /* unreachable anyway */
}
